
//========================================================================================
//
//		Fun with prime numbers - the sieve of Eratosthenes v3.1 
//		by Dan Brown
//		www.tacticaltinkering.co.uk
//
//========================================================================================

#include <iostream>
#include <Windows.h>

#define MAX			15485863		//15485863  //179424673 //982451653 //2038074743 // 22801763489
#define GetPos(n)	((n-3)/2)		// We're now only holding odd numbers in our array, so we need this macro to map each number to the char
									// which represents it - ignoring number 1, so the first element in the array corresponds to number 3

#if MAX > 4294967295				// If we can, use 32 bit integers (twice as fast on a 32 bit system)
typedef unsigned __int64 BigInt;
#else
typedef unsigned int BigInt;
#endif

using namespace std;

//-----------------------------------------------------------------------------------------
// Name: CTimer [CLASS]
// Desc: A class for use as a fairly accurate stopwatch
//-----------------------------------------------------------------------------------------
class CTimer	{
	private:
		__int64 start;
		__int64 end;
		__int64 iElapsed()
		{
			LARGE_INTEGER lInt;
			QueryPerformanceCounter(&lInt);
			return (lInt.QuadPart-start);
		}
	public:
		double value;
		double valueMS;
		bool Start()
		{
			LARGE_INTEGER lInt;
			if(!QueryPerformanceCounter(&lInt))
			{
				std::cout << "\nTimer failed to start!";
				return false;
			}
			start = lInt.QuadPart;
			return true;
		}
		bool Stop()
		{
			LARGE_INTEGER lIntPF, lInt;
			if(!QueryPerformanceFrequency(&lIntPF)) return false;
			if(!QueryPerformanceCounter(&lInt)) return false;			
			end = lInt.QuadPart-start;
			value = (double)iElapsed()/(double)lIntPF.QuadPart;
			valueMS = value*1000;
			return true;
		}
		double TimeElapsed()
		{
			Stop();
			return value;
		}
} Timer;
//-----------------------------------------------------------------------------------------

const BigInt arrSize = (MAX-1)/(2*32)+1;			// Number of "blocks of 32" elements needed for all odd numbers from 3 up to MAX
unsigned int *n = new unsigned int[arrSize];		// Create the main array on the heap

//----------------------------------------------------------------------------------------
// BitArray functions (for 32 bit ints)
// - ClearBit(n) - Sets the nth bit in the n[] array to 0
// - FlipBit(n)	 - Flips the nth bit in the n[] array
// - SetBit(n)	 - Sets the nth bit in the n[] array to 1
// - GetBit(n)	 - Retrieves the state of bit n
//				 -- Originally used "(n[bit/32] & (2147483648 >> bit)) > 0 ? 1 : 0",
//					but it turns out that shifting back to get 1 is up to 1% faster ...
//
//	EDIT: Using macros instead gives a 65% speed increase!
//----------------------------------------------------------------------------------------
//void ClearBit(BigInt bit)	{ n[(bit)>>5] &= ~(2147483648 >> bit); }	
//void FlipBit(BigInt bit)	{ n[(bit)>>5] ^= (2147483648 >> bit); }
//void SetBit(BigInt bit)		{ n[(bit)>>5] |= (2147483648 >> bit); }
//int GetBit(BigInt bit)		{ return (n[(bit)>>5] & (2147483648 >> bit)) >> (31-bit); }

#define ClearBit(bit)	n[(bit)>>5] &= ~(2147483648 >> bit)
#define FlipBit(bit)	n[(bit)>>5] ^= (2147483648 >> bit)
#define SetBit(bit)		n[(bit)>>5] |= (2147483648 >> bit)
#define GetBit(bit)		((n[(bit)>>5] & (2147483648 >> bit)) >> (31-bit))


int main()
{
	Timer.Start();

	std::fill(n, n + arrSize, 4294967295);	// Assume all odd numbers (bar 1) are prime to start, so set all array values to 1
											// 4294967295 = decimal form of 11111111111111111111111111111111

	cout << "\nInitialised the array (" << Timer.TimeElapsed() << "s)\n\nGenerating primes with the sieve of Eratosthenes ...\n\n";

	BigInt p = 3;			// p = current prime we're deleting multiples of in the bit array - ignoring 2!
	BigInt m = p*p;			// m = multiples of p greater than (p-1)*p, smaller multiples are already removed by smaller primes

	while(m <= MAX)
	{
		while(m <= MAX)							// Loop through the array and remove all multiples of p
		{
			ClearBit(GetPos(m));
			m += 2*p;							// Ignore even multiples
		}
		
		do {p+=2;} while( GetBit(GetPos(p)) != 1 );	// Let p be the next surviving number in the list
		m=p*p;									// Starting multiple of p
	}

	Timer.Stop();
	cout << "\n\tFinished sieving.\n\n\tTime taken:\t\t";
	Timer.value > 1 ? cout << Timer.value << "s" : cout << Timer.valueMS << "ms";
	cout << "\n\n\nCounting primes ...\n\n";
	
	Timer.Start();
	BigInt primesFound = 1;						// Not forgetting 2!
	for( BigInt a = 0 ; a < arrSize ; a++ )
		for( int b = 0 ; b < 32 ; b++ )
			if( (n[a] & (1 << (31-b&31))) > 0 )
			{
				if( (32*a+b)*2+3 > MAX ) break; // Don't include superflouous elements

				primesFound++; 

				if(primesFound==1000000) cout << "\n\tMillionth prime:\t" <<  (32*a+b)*2+3;
				if(primesFound==10000000) cout << "\n\t10 millionth prime:\t" <<  (32*a+b)*2+3;
				if(primesFound==50000000) cout << "\n\t50 millionth prime:\t" <<  (32*a+b)*2+3;
				if(primesFound==100000000) cout << "\n\t100 millionth prime:\t" <<  (32*a+b)*2+3;
				if(primesFound==1000000000) cout << "\n\tBillionth prime:\t" <<  (32*a+b)*2+3;
			}
	
	Timer.Stop();
	cout << "\n\n\tPrimes found:\t\t" << primesFound << "\n\tSearched up to:\t\t" << MAX << "\n\n\tCounted in:\t\t";
	Timer.value > 1 ? cout << Timer.value << "s" : cout << Timer.valueMS << "ms";
	
	cin.ignore();

	delete [] n;
}